@@ -1,5 +1,5 @@
 /* darkstat 3
- * copyright (c) 2001-2007 Emil Mikulic.
+ * copyright (c) 2001-2011 Emil Mikulic.
  *
  * conv.c: convenience functions.
  *
@@ -16,7 +16,6 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-#include "darkstat.h"
 #include "conv.h"
 
 #include <sys/wait.h>
@@ -31,6 +30,7 @@
 #include <string.h>
 #include <time.h>
 #include <unistd.h>
+#include <limits.h>
 
 #define PATH_DEVNULL "/dev/null"
 
@@ -135,9 +135,9 @@ str_starts_with(const char *haystack, const char *needle)
  *    num_chunks = 2, chunks = { "one", "two", NULL }
  */
 char **
-split(const char delimiter, const char *str, int *num_chunks)
+split(const char delimiter, const char *str, unsigned int *num_chunks)
 {
-   int num = 0;
+   unsigned int num = 0;
    char **chunks = NULL;
    size_t left, right = 0;
 
@@ -243,7 +243,8 @@ daemonize_start(void)
       verbosef("parent waiting");
       if (close(lifeline[1]) == -1)
          warn("close lifeline in parent");
-      read(lifeline[0], tmp, sizeof(tmp));
+      if (read(lifeline[0], tmp, sizeof(tmp)) != 0) /* expecting EOF */
+         err(1, "lifeline read() failed");
       verbosef("parent done reading, calling waitpid");
       w = waitpid(f, &status, WNOHANG);
       verbosef("waitpid ret %d, status is %d", w, status);
@@ -283,25 +284,37 @@ daemonize_finish(void)
       close(fd_null);
 }
 
-/* Security: chroot (optionally) and drop privileges. */
+/*
+ * For security, chroot (optionally) and drop privileges.
+ * Pass a NULL chroot_dir to disable chroot() behaviour.
+ */
 void
-privdrop(const int want_chroot)
+privdrop(const char *chroot_dir, const char *privdrop_user)
 {
-   struct passwd *pw = getpwnam(PRIVDROP_USER);
+   struct passwd *pw;
 
-   if (pw == NULL)
-      err(1, "getpwnam(" PRIVDROP_USER ")");
-   if (want_chroot) {
+   errno = 0;
+   pw = getpwnam(privdrop_user);
+
+   if (pw == NULL) {
+      if (errno == 0)
+         errx(1, "getpwnam(\"%s\") failed: no such user", privdrop_user);
+      else
+         err(1, "getpwnam(\"%s\") failed", privdrop_user);
+   }
+   if (chroot_dir != NULL) {
       tzset(); /* read /etc/localtime before we chroot */
-      if (chdir(CHROOT_DIR) == -1)
-         err(1, "chdir(" CHROOT_DIR ") failed");
-      if (chroot(CHROOT_DIR) == -1)
-         err(1, "chroot(" CHROOT_DIR ") failed");
+      if (chdir(chroot_dir) == -1)
+         err(1, "chdir(\"%s\") failed", chroot_dir);
+      if (chroot(chroot_dir) == -1)
+         err(1, "chroot(\"%s\") failed", chroot_dir);
+      verbosef("chrooted into: %s", chroot_dir);
    }
    if (setgid(pw->pw_gid) == -1)
       err(1, "setgid");
    if (setuid(pw->pw_uid) == -1)
       err(1, "setuid");
+   verbosef("set uid/gid to %d/%d", (int)pw->pw_uid, (int)pw->pw_gid);
 }
 
 /* Make the specified file descriptor non-blocking. */
@@ -331,116 +344,3 @@ fd_set_block(const int fd)
       err(1, "fcntl(fd %d) to unset O_NONBLOCK", fd);
    assert( (fcntl(fd, F_GETFL, 0) & O_NONBLOCK ) == 0 );
 }
-
-/* strlcpy() and strlcat() are derived from:
- *
- * $OpenBSD: strlcpy.c,v 1.4
- * $FreeBSD: src/lib/libc/string/strlcpy.c,v 1.8
- *
- * $OpenBSD: strlcat.c,v 1.2
- * $FreeBSD: src/lib/libc/string/strlcat.c,v 1.10
- *
- * under the following license:
- *
- * Copyright (c) 1998 Todd C. Miller <Todd.Miller@courtesan.com>
- * All rights reserved.
- *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in the
- *    documentation and/or other materials provided with the distribution.
- * 3. The name of the author may not be used to endorse or promote products
- *    derived from this software without specific prior written permission.
- *
- * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES,
- * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
- * AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL
- * THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL,
- * EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO,
- * PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS;
- * OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
- * WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
- * OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF
- * ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
- */
-
-#ifndef HAVE_STRLCPY
-/*
- * Copy src to string dst of size siz.  At most siz-1 characters
- * will be copied.  Always NUL terminates (unless siz == 0).
- * Returns strlen(src); if retval >= siz, truncation occurred.
- */
-size_t strlcpy(dst, src, siz)
-        char *dst;
-        const char *src;
-        size_t siz;
-{
-        char *d = dst;
-        const char *s = src;
-        size_t n = siz;
-
-        /* Copy as many bytes as will fit */
-        if (n != 0 && --n != 0) {
-                do {
-                        if ((*d++ = *s++) == 0)
-                                break;
-                } while (--n != 0);
-        }
-
-        /* Not enough room in dst, add NUL and traverse rest of src */
-        if (n == 0) {
-                if (siz != 0)
-                        *d = '\0';              /* NUL-terminate dst */
-                while (*s++)
-                        ;
-        }
-
-        return(s - src - 1);    /* count does not include NUL */
-}
-#endif
-
-#ifndef HAVE_STRLCAT
-/*
- * Appends src to string dst of size siz (unlike strncat, siz is the
- * full size of dst, not space left).  At most siz-1 characters
- * will be copied.  Always NUL terminates (unless siz <= strlen(dst)).
- * Returns strlen(src) + MIN(siz, strlen(initial dst)).
- * If retval >= siz, truncation occurred.
- */
-size_t
-strlcat(dst, src, siz)
-        char *dst;
-        const char *src;
-        size_t siz;
-{
-        char *d = dst;
-        const char *s = src;
-        size_t n = siz;
-        size_t dlen;
-
-        /* Find the end of dst and adjust bytes left but don't go past end */
-        while (n-- != 0 && *d != '\0')
-                d++;
-        dlen = d - dst;
-        n = siz - dlen;
-
-        if (n == 0)
-                return(dlen + strlen(s));
-        while (*s != '\0') {
-                if (n != 1) {
-                        *d++ = *s;
-                        n--;
-                }
-                s++;
-        }
-        *d = '\0';
-
-        return(dlen + (s - src));       /* count does not include NUL */
-}
-#endif
-
-/* vim:set ts=3 sw=3 tw=78 expandtab: */
