@@ -1,5 +1,5 @@
 /* darkstat 3
- * copyright (c) 2001-2008 Emil Mikulic.
+ * copyright (c) 2001-2011 Emil Mikulic.
  *
  * err.h: BSD-like err() and warn() functions
  *
@@ -16,14 +16,17 @@
  * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
  */
 
-void err(const int code, const char *format, ...);
-void errx(const int code, const char *format, ...);
+#include "cdefs.h"
 
-void warn(const char *format, ...);
-void warnx(const char *format, ...);
+void err(const int code, const char *format, ...)
+   _noreturn_ _printflike_(2, 3);
+void errx(const int code, const char *format, ...)
+   _noreturn_ _printflike_(2, 3);
 
-extern int want_verbose;
-void verbosef(const char *format, ...);
+void warn(const char *format, ...) _printflike_(1, 2);
+void warnx(const char *format, ...) _printflike_(1, 2);
+
+void verbosef(const char *format, ...) _printflike_(1, 2);
 void dverbosef(const char *format _unused_, ...);
 
 /* vim:set ts=3 sw=3 tw=78 expandtab: */
