@@ -1,5 +1,5 @@
 /* darkstat 3
- * copyright (c) 2006-2008 Emil Mikulic.
+ * copyright (c) 2006-2011 Emil Mikulic.
  *
  * graph_db.c: round robin database for graph data
  *
@@ -8,17 +8,17 @@
  */
 
 #include <sys/types.h>
-#include <netinet/in.h> /* for in_addr_t (db.h needs it) */
 
+#include "cap.h"
 #include "conv.h"
-#include "darkstat.h"
 #include "db.h"
 #include "acct.h"
 #include "err.h"
 #include "str.h"
-#include "html.h" /* FIXME: should be pushed into a .c file? */
+#include "html.h"
 #include "graph_db.h"
 #include "now.h"
+#include "opt.h"
 
 #include <assert.h>
 #include <stdlib.h>
@@ -27,8 +27,6 @@
 #define GRAPH_WIDTH "320"
 #define GRAPH_HEIGHT "200"
 
-extern const char *interface;
-
 struct graph {
    uint64_t *in, *out;
    unsigned int offset; /* i.e. seconds start at 0, days start at 1 */
@@ -301,43 +299,6 @@ graph_export(const int fd)
    return 1;
 }
 
-/* ---------------------------------------------------------------------------
- * Format a length of time in seconds to something more readable.
- */
-static struct str *
-length_of_time(const time_t t)
-{
-   struct str *buf = str_make();
-   int secs  =  t % 60;
-   int mins  = (t / 60) % 60;
-   int hours = (t / 3600) % 24;
-   int days  =  t / 86400;
-
-   int show_zeroes = 0;
-
-   if (days > 0) {
-      str_appendf(buf, "%d %s", days, (days==1)?"day":"days");
-      show_zeroes = 1;
-   }
-
-   if (show_zeroes || (hours > 0)) {
-      if (show_zeroes) str_append(buf, ", ");
-      str_appendf(buf, "%d %s", hours, (hours==1)?"hr":"hrs");
-      show_zeroes = 1;
-   }
-
-   if (show_zeroes || (mins > 0)) {
-      if (show_zeroes) str_append(buf, ", ");
-      str_appendf(buf, "%d %s", mins, (mins==1)?"min":"mins");
-      show_zeroes = 1;
-   }
-
-   if (show_zeroes) str_append(buf, ", ");
-   str_appendf(buf, "%d %s", secs, (secs==1)?"sec":"secs");
-
-   return buf;
-}
-
 /* ---------------------------------------------------------------------------
  * Web interface: front page!
  */
@@ -349,16 +310,9 @@ html_front_page(void)
    char start_when[100];
 
    buf = str_make();
-   str_append(buf, html_header_1);
-   str_appendf(buf, "<title>" PACKAGE_STRING " : graphs (%s)</title>\n",
-               interface);
-   str_append(buf, "<script src=\"graph.js\" type=\"text/javascript\">"
-                   "</script>\n");
-   str_append(buf, html_header_2);
-   str_appendf(buf, "<h2 class=\"pageheader\">Graphs (%s)</h2>\n", interface);
+   html_open(buf, "Graphs", /*path_depth=*/0, /*want_graph_js=*/1);
 
    str_append(buf, "<p>\n");
-
    str_append(buf, "<b>Running for</b> <span id=\"rf\">");
    rf = length_of_time(now - start_time);
    /* FIXME: use a more monotonic clock perhaps? */
@@ -370,10 +324,15 @@ html_front_page(void)
       "%Y-%m-%d %H:%M:%S %Z%z", localtime(&start_time)) != 0)
       str_appendf(buf, "<b>, since</b> %s", start_when);
 
-   str_appendf(buf,"<b>.</b><br/>\n"
-      "<b>Total</b> <span id=\"tb\">%'qu</span> <b>bytes in</b> "
-      "<span id=\"tp\">%'qu</span> <b>packets.</b><br/>\n</p>\n",
-      total_bytes, total_packets);
+   str_appendf(buf,"<b>.</b><br>\n"
+      "<b>Total</b> <span id=\"tb\">%'qu</span> <b>bytes, "
+      "in</b> <span id=\"tp\">%'qu</span> <b>packets.</b> "
+      "(<span id=\"pc\">%'u</span> <b>captured,</b> "
+      "<span id=\"pd\">%'u</span> <b>dropped)</b><br>\n"
+      "</p>\n",
+      acct_total_bytes,
+      acct_total_packets,
+      cap_pkts_recv, cap_pkts_drop);
 
    str_append(buf,
       "<div id=\"graphs\">\n"
@@ -383,7 +342,7 @@ html_front_page(void)
       "var graph_width = " GRAPH_WIDTH ";\n"
       "var graph_height = " GRAPH_HEIGHT ";\n"
       "var bar_gap = 1;\n"
-      "var graphs_uri = \"/graphs.xml\";\n"
+      "var graphs_uri = \"graphs.xml\";\n"
       "var graphs = [\n"
    );
 
@@ -406,7 +365,7 @@ html_front_page(void)
       "</div>\n"
    );
 
-   str_append(buf, html_footer);
+   html_close(buf);
    return (buf);
 }
 
@@ -419,8 +378,8 @@ xml_graphs(void)
    unsigned int i, j;
    struct str *buf = str_make(), *rf;
 
-   str_appendf(buf, "<graphs tp=\"%qu\" tb=\"%qu\" rf=\"",
-      total_packets, total_bytes);
+   str_appendf(buf, "<graphs tp=\"%qu\" tb=\"%qu\" pc=\"%u\" pd=\"%u\" rf=\"",
+      acct_total_packets, acct_total_bytes, cap_pkts_recv, cap_pkts_drop);
    rf = length_of_time(now - start_time);
    str_appendstr(buf, rf);
    str_free(rf);
