@@ -1,23 +1,25 @@
 /* darkstat 3
- * copyright (c) 2001-2008 Emil Mikulic.
+ * copyright (c) 2001-2011 Emil Mikulic.
  *
  * hosts_db.h: database of hosts, ports, protocols.
  *
  * You may use, modify and redistribute this file under the terms of the
  * GNU General Public License version 2. (see COPYING.GPL)
  */
+#ifndef __DARKSTAT_HOSTS_DB_H
+#define __DARKSTAT_HOSTS_DB_H
 
-#include <sys/types.h>
-#include <netinet/in.h>
-#include "str.h"
+#include <sys/types.h> /* for time_t and uint64_t (esp on FreeBSD) */
+
+#include "addr.h"
 
 struct hashtable;
 
 struct host {
-   in_addr_t ip;
+   struct addr addr;
    char *dns;
    uint8_t mac_addr[6];
-   time_t last_seen;
+   time_t lastseen;
    struct hashtable *ports_tcp, *ports_udp, *ip_protos;
 };
 
@@ -45,15 +47,9 @@ struct bucket {
    } u;
 };
 
-enum sort_dir { IN, OUT, TOTAL };
+enum sort_dir { IN, OUT, TOTAL, LASTSEEN };
 
-extern int show_mac_addrs;
-
-/*
- * Table reduction - when the number of entries is about to exceed <max>, we
- * reduce the table to the top <keep> entries.
- */
-extern unsigned int hosts_max, hosts_keep, ports_max, ports_keep;
+extern int hosts_db_show_macs;
 
 void hosts_db_init(void);
 void hosts_db_reduce(void);
@@ -62,8 +58,8 @@ void hosts_db_free(void);
 int hosts_db_import(const int fd);
 int hosts_db_export(const int fd);
 
-struct bucket *host_find(const in_addr_t ip); /* can return NULL */
-struct bucket *host_get(const in_addr_t ip);
+struct bucket *host_find(const struct addr *const a); /* can return NULL */
+struct bucket *host_get(const struct addr *const a);
 struct bucket *host_get_port_tcp(struct bucket *host, const uint16_t port);
 struct bucket *host_get_port_udp(struct bucket *host, const uint16_t port);
 struct bucket *host_get_ip_proto(struct bucket *host, const uint8_t proto);
@@ -75,4 +71,5 @@ struct str *html_hosts(const char *uri, const char *query);
 void qsort_buckets(const struct bucket **a, size_t n,
    size_t left, size_t right, const enum sort_dir d);
 
+#endif /* __DARKSTAT_HOSTS_DB_H */
 /* vim:set ts=3 sw=3 tw=78 expandtab: */
