@@ -1,5 +1,5 @@
 /* darkstat 3
- * copyright (c) 2001-2006 Emil Mikulic.
+ * copyright (c) 2001-2011 Emil Mikulic.
  *
  * ncache.c: cache of protocol and service names.
  *
@@ -7,11 +7,11 @@
  * GNU General Public License version 2. (see COPYING.GPL)
  */
 
-#include "darkstat.h"
 #include "conv.h"
 #include "err.h"
 #include "ncache.h"
 #include "tree.h"
+#include "bsd.h" /* for strlcpy */
 
 #include <netinet/in.h> /* ntohs */
 #include <netdb.h>
