@@ -1,5 +1,5 @@
 /* darkstat 3
- * copyright (c) 2001-2007 Emil Mikulic.
+ * copyright (c) 2001-2011 Emil Mikulic.
  *
  * str.h: string buffer with pool-based reallocation
  *
@@ -19,20 +19,20 @@
 #include <sys/types.h>
 #include <stdarg.h>
 
-/* Note: the contents are 8-bit clean but not zero terminated! */
+/* Note: the contents are 8-bit clean and not zero terminated! */
 
 struct str;
 
 struct str *str_make(void);
 void str_free(struct str *s);
 void str_extract(struct str *buf, size_t *len, char **str);
-void str_append2(struct str *buf, const char *s, const size_t len);
+void str_appendn(struct str *buf, const char *s, const size_t len);
 void str_appendstr(struct str *buf, const struct str *s);
 
 #ifdef __GNUC__
 /* amusing efficiency hack */
 #include <string.h>
-#define str_append(buf, s) str_append2(buf, s, \
+#define str_append(buf, s) str_appendn(buf, s, \
     (__builtin_constant_p(s) ? sizeof(s)-1 : strlen(s)) )
 #else
 void str_append(struct str *buf, const char *s);
@@ -42,4 +42,6 @@ size_t xvasprintf(char **result, const char *format, va_list va);
 size_t xasprintf(char **result, const char *format, ...);
 void str_appendf(struct str *buf, const char *format, ...);
 
+struct str *length_of_time(const time_t t);
+
 /* vim:set ts=3 sw=3 tw=78 expandtab: */
